import copy
import asyncio

from typing import Union, Type, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from edsl.questions.QuestionBase import QuestionBase

from edsl.surveys.base import EndOfSurvey
from edsl.jobs.tasks.task_status_enum import TaskStatus

from edsl.jobs.FetchInvigilator import FetchInvigilator
from edsl.exceptions.language_models import LanguageModelNoResponseError
from edsl.exceptions.questions import QuestionAnswerValidationError
from edsl.data_transfer_models import AgentResponseDict, EDSLResultObjectInput

from edsl.jobs.Answers import Answers


class AnswerQuestionFunctionConstructor:
    def __init__(self, interview):
        self.interview = interview
        self.had_language_model_no_response_error = False
        self.question_index = self.interview.to_index

        self.skip_function: Callable = (
            self.interview.survey.rule_collection.skip_question_before_running
        )

    def _combined_answers(self) -> Answers:
        return self.answers | self.interview.scenario | self.interview.agent["traits"]

    @property
    def answers(self) -> Answers:
        return self.interview.answers

    def _skip_this_question(self, current_question: "QuestionBase") -> bool:
        current_question_index = self.question_index[current_question.question_name]
        combined_answers = self._combined_answers()
        return self.skip_function(current_question_index, combined_answers)

    def _handle_exception(
        self, e: Exception, invigilator: "InvigilatorBase", task=None
    ):
        from edsl.jobs.interviews.InterviewExceptionEntry import InterviewExceptionEntry

        answers = copy.copy(self.answers)  # copy to freeze the answers here for logging
        exception_entry = InterviewExceptionEntry(
            exception=e,
            invigilator=invigilator,
            answers=answers,
        )
        if task:
            task.task_status = TaskStatus.FAILED
        self.interview.exceptions.add(
            invigilator.question.question_name, exception_entry
        )

        if self.interview.raise_validation_errors and isinstance(
            e, QuestionAnswerValidationError
        ):
            raise e

        stop_on_exception = getattr(self.interview, "stop_on_exception", False)
        if stop_on_exception:
            raise e

    def _cancel_skipped_questions(self, current_question: "QuestionBase") -> None:
        current_question_index: int = self.interview.to_index[
            current_question.question_name
        ]
        answers = (
            self.answers | self.interview.scenario | self.interview.agent["traits"]
        )

        # Get the index of the next question, which could also be the end of the survey
        next_question: Union[
            int, EndOfSurvey
        ] = self.interview.survey.rule_collection.next_question(
            q_now=current_question_index,
            answers=answers,
        )

        def cancel_between(start, end):
            for i in range(start, end):
                self.interview.tasks[i].cancel()

        if (next_question_index := next_question.next_q) == EndOfSurvey:
            cancel_between(
                current_question_index + 1, len(self.interview.survey.questions)
            )
            return

        if next_question_index > (current_question_index + 1):
            cancel_between(current_question_index + 1, next_question_index)

    def __call__(self):
        from edsl.config import CONFIG

        EDSL_BACKOFF_START_SEC = float(CONFIG.get("EDSL_BACKOFF_START_SEC"))
        EDSL_BACKOFF_MAX_SEC = float(CONFIG.get("EDSL_BACKOFF_MAX_SEC"))
        EDSL_MAX_ATTEMPTS = int(CONFIG.get("EDSL_MAX_ATTEMPTS"))

        from tenacity import (
            retry,
            stop_after_attempt,
            wait_exponential,
            retry_if_exception_type,
            RetryError,
        )

        async def answer_question_and_record_task(
            *,
            question: "QuestionBase",
            task=None,
        ) -> "AgentResponseDict":
            @retry(
                stop=stop_after_attempt(EDSL_MAX_ATTEMPTS),
                wait=wait_exponential(
                    multiplier=EDSL_BACKOFF_START_SEC, max=EDSL_BACKOFF_MAX_SEC
                ),
                retry=retry_if_exception_type(LanguageModelNoResponseError),
                reraise=True,
            )
            async def attempt_answer():
                invigilator = FetchInvigilator(self.interview)(question)

                if self._skip_this_question(question):
                    return invigilator.get_failed_task_result(
                        failure_reason="Question skipped."
                    )

                try:
                    response: EDSLResultObjectInput = (
                        await invigilator.async_answer_question()
                    )
                    if response.validated:
                        self.answers.add_answer(response=response, question=question)
                        self._cancel_skipped_questions(question)
                    else:
                        if (
                            hasattr(response, "exception_occurred")
                            and response.exception_occurred
                        ):
                            raise response.exception_occurred

                except QuestionAnswerValidationError as e:
                    self._handle_exception(e, invigilator, task)
                    return invigilator.get_failed_task_result(
                        failure_reason="Question answer validation failed."
                    )

                except asyncio.TimeoutError as e:
                    self._handle_exception(e, invigilator, task)
                    had_language_model_no_response_error = True
                    raise LanguageModelNoResponseError(
                        f"Language model timed out for question '{question.question_name}.'"
                    )

                except Exception as e:
                    self._handle_exception(e, invigilator, task)

                if "response" not in locals():
                    had_language_model_no_response_error = True
                    raise LanguageModelNoResponseError(
                        f"Language model did not return a response for question '{question.question_name}.'"
                    )

                if (
                    question.question_name in self.interview.exceptions
                    and had_language_model_no_response_error
                ):
                    self.interview.exceptions.record_fixed_question(
                        question.question_name
                    )

                return response

            try:
                return await attempt_answer()
            except RetryError as retry_error:
                original_error = retry_error.last_attempt.exception()
                self._handle_exception(
                    original_error, FetchInvigilator(self.interview)(question), task
                )
                raise original_error

        return answer_question_and_record_task



from typing import Optional, Union, Literal, TYPE_CHECKING, NewType


Seconds = NewType("Seconds", float)
JobUUID = NewType("JobUUID", str)

from edsl.exceptions.coop import CoopServerResponseError

if TYPE_CHECKING:
    from edsl.results.Results import Results
    from edsl.jobs.Jobs import Jobs
    from edsl.coop.coop import RemoteInferenceResponse, RemoteInferenceCreationInfo
    from edsl.jobs.JobsRemoteInferenceLogger import JobLogger

from edsl.coop.coop import RemoteInferenceResponse, RemoteInferenceCreationInfo

from edsl.jobs.jobs_status_enums import JobsStatus
from edsl.coop.utils import VisibilityType


class JobsRemoteInferenceHandler:
    def __init__(self, jobs: "Jobs", verbose: bool = False, poll_interval: Seconds = 1):
        """ """
        self.jobs = jobs
        self.verbose = verbose
        self.poll_interval = poll_interval

        self._remote_job_creation_data: Union[None, RemoteInferenceCreationInfo] = None
        self._job_uuid: Union[None, JobUUID] = None  # Will be set when job is created
        self.logger: Union[None, JobLogger] = None  # Will be initialized when needed

    @property
    def remote_job_creation_data(self) -> RemoteInferenceCreationInfo:
        return self._remote_job_creation_data

    @property
    def job_uuid(self) -> JobUUID:
        return self._job_uuid

    def use_remote_inference(self, disable_remote_inference: bool) -> bool:
        import requests

        if disable_remote_inference:
            return False
        if not disable_remote_inference:
            try:
                from edsl.coop.coop import Coop

                user_edsl_settings = Coop().edsl_settings
                return user_edsl_settings.get("remote_inference", False)
            except requests.ConnectionError:
                pass
            except CoopServerResponseError as e:
                pass

        return False

    def create_remote_inference_job(
        self,
        iterations: int = 1,
        remote_inference_description: Optional[str] = None,
        remote_inference_results_visibility: Optional[VisibilityType] = "unlisted",
    ) -> None:
        from edsl.config import CONFIG
        from edsl.coop.coop import Coop

        # Initialize logger
        from edsl.utilities.is_notebook import is_notebook
        from edsl.jobs.JobsRemoteInferenceLogger import JupyterJobLogger
        from edsl.jobs.JobsRemoteInferenceLogger import StdOutJobLogger
        from edsl.jobs.loggers.HTMLTableJobLogger import HTMLTableJobLogger

        if is_notebook():
            self.logger = HTMLTableJobLogger(verbose=self.verbose)
        else:
            self.logger = StdOutJobLogger(verbose=self.verbose)

        coop = Coop()
        self.logger.update(
            "Remote inference activated. Sending job to server...",
            status=JobsStatus.QUEUED,
        )
        remote_job_creation_data = coop.remote_inference_create(
            self.jobs,
            description=remote_inference_description,
            status="queued",
            iterations=iterations,
            initial_results_visibility=remote_inference_results_visibility,
        )
        self.logger.update(
            "Your survey is running at the Expected Parrot server...",
            status=JobsStatus.RUNNING,
        )

        job_uuid = remote_job_creation_data.get("uuid")
        self.logger.update(
            message=f"Job sent to server. (Job uuid={job_uuid}).",
            status=JobsStatus.RUNNING,
        )
        self.logger.add_info("job_uuid", job_uuid)

        expected_parrot_url = CONFIG.get("EXPECTED_PARROT_URL")
        remote_inference_url = f"{expected_parrot_url}/home/remote-inference"

        self.logger.update(
            f"Job details are available at your Coop account {remote_inference_url}{remote_inference_url}",
            status=JobsStatus.RUNNING,
        )
        progress_bar_url = f"{expected_parrot_url}/home/remote-job-progress/{job_uuid}"
        self.logger.add_info("progress_bar_url", progress_bar_url)
        self.logger.update(
            f"View job progress here: {progress_bar_url}", status=JobsStatus.RUNNING
        )

        self._remote_job_creation_data = remote_job_creation_data
        self._job_uuid = job_uuid

    @staticmethod
    def check_status(
        job_uuid: JobUUID,
    ) -> RemoteInferenceResponse:
        from edsl.coop.coop import Coop

        coop = Coop()
        return coop.remote_inference_get(job_uuid)

    def poll_remote_inference_job(self) -> Union[None, "Results"]:
        return self._poll_remote_inference_job(
            self.remote_job_creation_data, verbose=self.verbose
        )

    def _poll_remote_inference_job(
        self,
        remote_job_creation_data: RemoteInferenceCreationInfo,
        verbose: bool = False,
        poll_interval: Optional[Seconds] = None,
        testing_simulated_response=None,
    ) -> Union[None, "Results"]:
        import time
        from datetime import datetime
        from edsl.config import CONFIG
        from edsl.results.Results import Results

        if poll_interval is None:
            poll_interval = self.poll_interval

        job_uuid = remote_job_creation_data.get("uuid")
        expected_parrot_url = CONFIG.get("EXPECTED_PARROT_URL")

        if testing_simulated_response is not None:
            remote_job_data_fetcher = lambda job_uuid: testing_simulated_response
            object_fetcher = (
                lambda results_uuid, expected_object_type: Results.example()
            )
        else:
            from edsl.coop.coop import Coop

            coop = Coop()
            remote_job_data_fetcher = coop.remote_inference_get
            object_fetcher = coop.get

        job_in_queue = True
        while job_in_queue:
            remote_job_data: RemoteInferenceResponse = remote_job_data_fetcher(job_uuid)
            status = remote_job_data.get("status")

            if status == "cancelled":
                self.logger.update(
                    messaged="Job cancelled by the user.", status=JobsStatus.CANCELLED
                )
                self.logger.update(
                    f"See {expected_parrot_url}/home/remote-inference for more details.",
                    status=JobsStatus.CANCELLED,
                )
                return None

            elif status == "failed":
                latest_error_report_url = remote_job_data.get("latest_error_report_url")
                if latest_error_report_url:
                    self.logger.update("Job failed.", status=JobsStatus.FAILED)
                    self.logger.update(
                        f"Error report: {latest_error_report_url}", "failed"
                    )
                    self.logger.add_info("error_report_url", latest_error_report_url)
                    self.logger.update(
                        "Need support? Visit Discord: https://discord.com/invite/mxAYkjfy9m",
                        status=JobsStatus.FAILED,
                    )
                else:
                    self.logger.update("Job failed.", "failed")
                    self.logger.update(
                        f"See {expected_parrot_url}/home/remote-inference for details.",
                        status=JobsStatus.FAILED,
                    )

                results_uuid = remote_job_data.get("results_uuid")
                if results_uuid:
                    self.logger.add_info("results_uuid", results_uuid)
                    results = object_fetcher(
                        results_uuid, expected_object_type="results"
                    )
                    results.job_uuid = job_uuid
                    results.results_uuid = results_uuid
                    return results
                else:
                    return None

            elif status == "completed":
                results_uuid = remote_job_data.get("results_uuid")
                self.logger.add_info("results_uuid", results_uuid)
                results_url = remote_job_data.get("results_url")
                self.logger.add_info("results_url", results_url)
                results = object_fetcher(results_uuid, expected_object_type="results")
                self.logger.update(
                    f"Job completed and Results stored on Coop: {results_url}",
                    status=JobsStatus.COMPLETED,
                )
                results.job_uuid = job_uuid
                results.results_uuid = results_uuid
                return results

            else:
                time_checked = datetime.now().strftime("%Y-%m-%d %I:%M:%S %p")
                self.logger.update(
                    f"Job status: {status} - last update: {time_checked}",
                    status=JobsStatus.RUNNING,
                )
                time.sleep(poll_interval)

    def use_remote_inference(self, disable_remote_inference: bool) -> bool:
        import requests

        if disable_remote_inference:
            return False
        if not disable_remote_inference:
            try:
                from edsl.coop.coop import Coop

                user_edsl_settings = Coop().edsl_settings
                return user_edsl_settings.get("remote_inference", False)
            except requests.ConnectionError:
                pass
            except CoopServerResponseError as e:
                pass

        return False

    async def create_and_poll_remote_job(
        self,
        iterations: int = 1,
        remote_inference_description: Optional[str] = None,
        remote_inference_results_visibility: Optional[VisibilityType] = "unlisted",
    ) -> Union["Results", None]:
        """
        Creates and polls a remote inference job asynchronously.
        Reuses existing synchronous methods but runs them in an async context.

        :param iterations: Number of times to run each interview
        :param remote_inference_description: Optional description for the remote job
        :param remote_inference_results_visibility: Visibility setting for results
        :return: Results object if successful, None if job fails or is cancelled
        """
        import asyncio
        from functools import partial

        # Create job using existing method
        loop = asyncio.get_event_loop()
        remote_job_creation_data = await loop.run_in_executor(
            None,
            partial(
                self.create_remote_inference_job,
                iterations=iterations,
                remote_inference_description=remote_inference_description,
                remote_inference_results_visibility=remote_inference_results_visibility,
            ),
        )

        # Poll using existing method but with async sleep
        return await loop.run_in_executor(
            None, partial(self.poll_remote_inference_job, remote_job_creation_data)
        )


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from __future__ import annotations

import os
import time
import requests
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass

from typing import Any, List, DefaultDict, Optional, Dict
from collections import defaultdict
from uuid import UUID

from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage

InterviewTokenUsageMapping = DefaultDict[str, InterviewTokenUsage]

from edsl.jobs.interviews.InterviewStatistic import InterviewStatistic
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage


@dataclass
class ModelInfo:
    model_name: str
    TPM_limit_k: float
    RPM_limit_k: float
    num_tasks_waiting: int
    token_usage_info: dict


@dataclass
class ModelTokenUsageStats:
    token_usage_type: str
    details: List[dict]
    cost: str


class JobsRunnerStatusBase(ABC):
    def __init__(
        self,
        jobs_runner: "JobsRunnerAsyncio",
        n: int,
        refresh_rate: float = 1,
        endpoint_url: Optional[str] = "http://localhost:8000",
        job_uuid: Optional[UUID] = None,
        api_key: str = None,
    ):
        self.jobs_runner = jobs_runner

        # The uuid of the job on Coop
        self.job_uuid = job_uuid

        self.base_url = f"{endpoint_url}"

        self.start_time = time.time()
        self.completed_interviews = []
        self.refresh_rate = refresh_rate
        self.statistics = [
            "elapsed_time",
            "total_interviews_requested",
            "completed_interviews",
            #            "percent_complete",
            "average_time_per_interview",
            #            "task_remaining",
            "estimated_time_remaining",
            "exceptions",
            "unfixed_exceptions",
            "throughput",
        ]
        self.num_total_interviews = n * len(self.jobs_runner.interviews)

        self.distinct_models = list(
            set(i.model.model for i in self.jobs_runner.interviews)
        )

        self.completed_interview_by_model = defaultdict(list)

        self.api_key = api_key or os.getenv("EXPECTED_PARROT_API_KEY")

    @abstractmethod
    def has_ep_api_key(self):
        """
        Checks if the user has an Expected Parrot API key.
        """
        pass

    def get_status_dict(self) -> Dict[str, Any]:
        """
        Converts current status into a JSON-serializable dictionary.
        """
        # Get all statistics
        stats = {}
        for stat_name in self.statistics:
            stat = self._compute_statistic(stat_name)
            name, value = list(stat.items())[0]
            stats[name] = value

        # Calculate overall progress
        total_interviews = len(self.jobs_runner.total_interviews)
        completed = len(self.completed_interviews)

        # Get model-specific progress
        model_progress = {}
        for model in self.distinct_models:
            completed_for_model = len(self.completed_interview_by_model[model])
            target_for_model = int(
                self.num_total_interviews / len(self.distinct_models)
            )
            model_progress[model] = {
                "completed": completed_for_model,
                "total": target_for_model,
                "percent": (
                    (completed_for_model / target_for_model * 100)
                    if target_for_model > 0
                    else 0
                ),
            }

        status_dict = {
            "overall_progress": {
                "completed": completed,
                "total": total_interviews,
                "percent": (
                    (completed / total_interviews * 100) if total_interviews > 0 else 0
                ),
            },
            "language_model_progress": model_progress,
            "statistics": stats,
            "status": "completed" if completed >= total_interviews else "running",
        }

        model_queues = {}
        for model, bucket in self.jobs_runner.bucket_collection.items():
            model_name = model.model
            model_queues[model_name] = {
                "language_model_name": model_name,
                "requests_bucket": {
                    "completed": bucket.requests_bucket.num_released,
                    "requested": bucket.requests_bucket.num_requests,
                    "tokens_returned": bucket.requests_bucket.tokens_returned,
                    "target_rate": round(bucket.requests_bucket.target_rate, 1),
                    "current_rate": round(bucket.requests_bucket.get_throughput(), 1),
                },
                "tokens_bucket": {
                    "completed": bucket.tokens_bucket.num_released,
                    "requested": bucket.tokens_bucket.num_requests,
                    "tokens_returned": bucket.tokens_bucket.tokens_returned,
                    "target_rate": round(bucket.tokens_bucket.target_rate, 1),
                    "current_rate": round(bucket.tokens_bucket.get_throughput(), 1),
                },
            }
        status_dict["language_model_queues"] = model_queues
        return status_dict

    @abstractmethod
    def setup(self):
        """
        Conducts any setup that needs to happen prior to sending status updates.

        Ex. For a local job, creates a job in the Coop database.
        """
        pass

    @abstractmethod
    def send_status_update(self):
        """
        Updates the current status of the job.
        """
        pass

    def add_completed_interview(self, result):
        self.completed_interviews.append(result.interview_hash)

        relevant_model = result.model.model
        self.completed_interview_by_model[relevant_model].append(result.interview_hash)

    def _compute_statistic(self, stat_name: str):
        completed_tasks = self.completed_interviews
        elapsed_time = time.time() - self.start_time
        interviews = self.jobs_runner.total_interviews

        stat_definitions = {
            "elapsed_time": lambda: InterviewStatistic(
                "elapsed_time", value=elapsed_time, digits=1, units="sec."
            ),
            "total_interviews_requested": lambda: InterviewStatistic(
                "total_interviews_requested", value=len(interviews), units=""
            ),
            "completed_interviews": lambda: InterviewStatistic(
                "completed_interviews", value=len(completed_tasks), units=""
            ),
            "percent_complete": lambda: InterviewStatistic(
                "percent_complete",
                value=(
                    len(completed_tasks) / len(interviews) * 100
                    if len(interviews) > 0
                    else 0
                ),
                digits=1,
                units="%",
            ),
            "average_time_per_interview": lambda: InterviewStatistic(
                "average_time_per_interview",
                value=elapsed_time / len(completed_tasks) if completed_tasks else 0,
                digits=2,
                units="sec.",
            ),
            "task_remaining": lambda: InterviewStatistic(
                "task_remaining", value=len(interviews) - len(completed_tasks), units=""
            ),
            "estimated_time_remaining": lambda: InterviewStatistic(
                "estimated_time_remaining",
                value=(
                    (len(interviews) - len(completed_tasks))
                    * (elapsed_time / len(completed_tasks))
                    if len(completed_tasks) > 0
                    else 0
                ),
                digits=1,
                units="sec.",
            ),
            "exceptions": lambda: InterviewStatistic(
                "exceptions",
                value=sum(len(i.exceptions) for i in interviews),
                units="",
            ),
            "unfixed_exceptions": lambda: InterviewStatistic(
                "unfixed_exceptions",
                value=sum(i.exceptions.num_unfixed() for i in interviews),
                units="",
            ),
            "throughput": lambda: InterviewStatistic(
                "throughput",
                value=len(completed_tasks) / elapsed_time if elapsed_time > 0 else 0,
                digits=2,
                units="interviews/sec.",
            ),
        }
        return stat_definitions[stat_name]()

    def update_progress(self, stop_event):
        while not stop_event.is_set():
            self.send_status_update()
            time.sleep(self.refresh_rate)

        self.send_status_update()


class JobsRunnerStatus(JobsRunnerStatusBase):
    @property
    def create_url(self) -> str:
        return f"{self.base_url}/api/v0/local-job"

    @property
    def viewing_url(self) -> str:
        return f"{self.base_url}/home/local-job-progress/{str(self.job_uuid)}"

    @property
    def update_url(self) -> str:
        return f"{self.base_url}/api/v0/local-job/{str(self.job_uuid)}"

    def setup(self) -> None:
        """
        Creates a local job on Coop if one does not already exist.
        """

        headers = {"Content-Type": "application/json"}

        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        else:
            headers["Authorization"] = f"Bearer None"

        if self.job_uuid is None:
            # Create a new local job
            response = requests.post(
                self.create_url,
                headers=headers,
                timeout=1,
            )
        response.raise_for_status()
        data = response.json()
        self.job_uuid = data.get("job_uuid")

        print(f"Running with progress bar. View progress at {self.viewing_url}")

    def send_status_update(self) -> None:
        """
        Sends current status to the web endpoint using the instance's job_uuid.
        """
        try:
            # Get the status dictionary and add the job_id
            status_dict = self.get_status_dict()

            # Make the UUID JSON serializable
            status_dict["job_id"] = str(self.job_uuid)

            headers = {"Content-Type": "application/json"}

            if self.api_key:
                headers["Authorization"] = f"Bearer {self.api_key}"
            else:
                headers["Authorization"] = f"Bearer None"

            # Send the update
            response = requests.patch(
                self.update_url,
                json=status_dict,
                headers=headers,
                timeout=1,
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            print(f"Failed to send status update for job {self.job_uuid}: {e}")

    def has_ep_api_key(self) -> bool:
        """
        Returns True if the user has an Expected Parrot API key. Otherwise, returns False.
        """

        if self.api_key is not None:
            return True
        else:
            return False


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from collections import UserDict, defaultdict, UserList
from typing import Union
from edsl.enums import InferenceServiceLiteral
from dataclasses import dataclass


@dataclass
class LanguageModelInfo:
    model_name: str
    service_name: str

    def __getitem__(self, key: int) -> str:
        import warnings

        warnings.warn(
            "Accessing LanguageModelInfo via index is deprecated. "
            "Please use .model_name, .service_name, or .index attributes instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if key == 0:
            return self.model_name
        elif key == 1:
            return self.service_name
        else:
            raise IndexError("Index out of range")


class ModelNamesList(UserList):
    pass


class AvailableModels(UserList):
    def __init__(self, data: list) -> None:
        super().__init__(data)

    def __contains__(self, model_name: str) -> bool:
        for model_entry in self:
            if model_entry.model_name == model_name:
                return True
        return False


class ServiceToModelsMapping(UserDict):
    def __init__(self, data: dict) -> None:
        super().__init__(data)

    @property
    def service_names(self) -> list[str]:
        return list(self.data.keys())

    def _validate_service_names(self):
        for service in self.service_names:
            if service not in InferenceServiceLiteral:
                raise ValueError(f"Invalid service name: {service}")

    def model_to_services(self) -> dict:
        self._model_to_service = defaultdict(list)
        for service, models in self.data.items():
            for model in models:
                self._model_to_service[model].append(service)



# Schemas
from edsl.questions.settings import Settings
from edsl.questions.RegisterQuestionsMeta import RegisterQuestionsMeta

# Base Class
from edsl.questions.QuestionBase import QuestionBase

# Core Questions
from edsl.questions.QuestionCheckBox import QuestionCheckBox
from edsl.questions.QuestionExtract import QuestionExtract
from edsl.questions.QuestionFreeText import QuestionFreeText
from edsl.questions.QuestionFunctional import QuestionFunctional
from edsl.questions.QuestionList import QuestionList
from edsl.questions.QuestionMatrix import QuestionMatrix
from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
from edsl.questions.QuestionNumerical import QuestionNumerical
from edsl.questions.QuestionBudget import QuestionBudget
from edsl.questions.QuestionRank import QuestionRank

# # # Questions derived from core questions
from edsl.questions.derived.QuestionLikertFive import QuestionLikertFive
from edsl.questions.derived.QuestionLinearScale import QuestionLinearScale
from edsl.questions.derived.QuestionYesNo import QuestionYesNo
from edsl.questions.derived.QuestionTopK import QuestionTopK

# # Compose Questions
# from edsl.questions.compose_questions import compose_questions



class ResponseValidatorFactory:
    def __init__(self, question):
        self.question = question

    @property
    def response_model(self) -> type["BaseModel"]:
        if self.question._response_model is not None:
            return self.question._response_model
        else:
            return self.question.create_response_model()

    @property
    def response_validator(self) -> "ResponseValidatorBase":
        """Return the response validator."""
        params = (
            {
                "response_model": self.question.response_model,
            }
            | {k: getattr(self.question, k) for k in self.validator_parameters}
            | {"exception_to_throw": getattr(self.question, "exception_to_throw", None)}
            | {"override_answer": getattr(self.question, "override_answer", None)}
        )
        return self.question.response_validator_class(**params)

    @property
    def validator_parameters(self) -> list[str]:
        """Return the parameters required for the response validator."""
        return self.question.response_validator_class.required_params



def is_notebook() -> bool:
    """Check if the code is running in a Jupyter notebook or Google Colab."""
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "Shell":  # Google Colab's shell class
            import sys

            if "google.colab" in sys.modules:
                return True  # Running in Google Colab
            return False
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type
    except NameError:
        return False  # Probably standard Python interpreter
